@@ -1,48 +1,75 @@
-from typing import Any, Callable, Dict, List, Optional, Union
+from dataclasses import dataclass, field
+from typing import Callable, Dict, List, Optional, TypeVar
 
 import numpy as np
-from pydantic import Field, parse_obj_as, parse_raw_as, validate_arguments
+from apischema import schema
+from typing_extensions import Annotated
 
 from .core import (
     Dimension,
+    Midpoints,
     Path,
-    SpecPositions,
-    _WithType,
+    Serializable,
+    alternative_constructor,
     if_instance_do,
     squash_dimensions,
 )
 from .regions import Region, get_mask
 
+T = TypeVar("T")
 
-class Spec(_WithType):
-    """Abstract baseclass for the specification of a scan. Supports operators:
+__all__ = [
+    "Spec",
+    "Product",
+    "Zip",
+    "Mask",
+    "Snake",
+    "Concat",
+    "Squash",
+    "Line",
+    "Static",
+    "Spiral",
+    "TIME",
+    "REPEAT",
+    "fly",
+    "step",
+    "repeat",
+]
+
+
+@dataclass
+class Spec(Serializable):
+    """Definition: A spec is a serializable representation of the type, parameters
+    and axis names required to produce one or more dimensions.
+
+    Description: Abstract baseclass for the specification of a scan. Supports operators:
 
     - ``*``: Outer `Product` of two Specs, nesting the second within the first
     - ``+``: `Zip` two Specs together, iterating in tandem
-    - ``&``: `Mask` the Spec with a `Region`, excluding positions outside it
+    - ``&``: `Mask` the Spec with a `Region`, excluding midpoints outside of it
     - ``~``: `Snake` the Spec, reversing every other iteration of it
     """
 
-    def keys(self) -> List:
-        """Implemented by subclasses to produce the list of keys that are
-        present in the positions, from slowest moving to fastest moving"""
+    def axes(self) -> List:
+        """Return the list of axes that are present in the scan, from
+        slowest moving to fastest moving"""
         raise NotImplementedError(self)
 
     def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
         """Implemented by subclasses to produce the `Dimension` list that
-        contribute to positions, from slowest moving to fastest moving"""
+        contribute to midpoints, from slowest moving to fastest moving"""
         raise NotImplementedError(self)
 
     def path(self) -> Path:
         """Return a `Path` through the scan that can be consumed in chunks
-        to give positions and bounds"""
+        to give midpoints and bounds"""
         path = Path(self.create_dimensions())
         return path
 
-    def positions(self) -> SpecPositions:
-        """Return a `SpecPositions` that can be iterated position by position"""
-        sp = SpecPositions(self.create_dimensions(bounds=False))
-        return sp
+    def midpoints(self) -> Midpoints:
+        """Return `Midpoints` that can be iterated point by point"""
+        mp = Midpoints(self.create_dimensions(bounds=False))
+        return mp
 
     def __mul__(self, other) -> "Product":
         return if_instance_do(other, Spec, lambda o: Product(self, o))
@@ -57,6 +84,7 @@ class Spec(_WithType):
         return Snake(self)
 
 
+@dataclass
 class Product(Spec):
     """Outer product of two Specs, nesting inner within outer. This means that
     inner will run in its entirety at each point in outer.
@@ -68,11 +96,13 @@ class Product(Spec):
         spec = Line("y", 1, 2, 3) * Line("x", 3, 4, 12)
     """
 
-    outer: Spec = Field(..., description="Will be executed once")
-    inner: Spec = Field(..., description="Will be executed len(outer) times")
+    outer: Spec = field(metadata=schema(description="Will be executed once"))
+    inner: Spec = field(
+        metadata=schema(description="Will be executed len(outer) times")
+    )
 
-    def keys(self) -> List:
-        return self.outer.keys() + self.inner.keys()
+    def axes(self) -> List:
+        return self.outer.axes() + self.inner.axes()
 
     def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
         dims_outer = self.outer.create_dimensions(bounds=False, nested=nested)
@@ -80,8 +110,9 @@ class Product(Spec):
         return dims_outer + dims_inner
 
 
+@dataclass
 class Zip(Spec):
-    """Run two Specs in parallel, merging their positions together. Typically
+    """Run two Specs in parallel, merging their midpoints together. Typically
     formed using the ``+`` operator.
 
     Dimensions are merged by:
@@ -101,15 +132,19 @@ class Zip(Spec):
         spec = Line("z", 1, 2, 3) * Line("y", 3, 4, 5) + Line("x", 4, 5, 5)
     """
 
-    left: Spec = Field(
-        ..., description="The left-hand Spec to Zip, will appear earlier in keys"
+    left: Spec = field(
+        metadata=schema(
+            description="The left-hand Spec to Zip, will appear earlier in axes"
+        )
     )
-    right: Spec = Field(
-        ..., description="The right-hand Spec to Zip, will appear later in keys"
+    right: Spec = field(
+        metadata=schema(
+            description="The right-hand Spec to Zip, will appear later in axes"
+        )
     )
 
-    def keys(self) -> List:
-        return self.left.keys() + self.right.keys()
+    def axes(self) -> List:
+        return self.left.axes() + self.right.axes()
 
     def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
         dims_left = self.left.create_dimensions(bounds, nested)
@@ -144,8 +179,9 @@ class Zip(Spec):
         return dimensions
 
 
+@dataclass
 class Mask(Spec):
-    """Restrict the given Spec to only the positions that fall inside the given
+    """Restrict the given Spec to only the midpoints that fall inside of the given
     Region.
 
     Typically created with the ``&`` operator. It also pushes down the
@@ -166,31 +202,38 @@ class Mask(Spec):
         spec = Line("y", 1, 3, 3) * Line("x", 3, 5, 5) & Circle("x", "y", 4, 2, 1.2)
     """
 
-    spec: Spec = Field(..., description="The Spec containing the source positions")
-    region: Region = Field(..., description="The Region that positions will be inside")
-    check_path_changes: bool = Field(
-        True, description="If True path through scan will not be modified by squash"
+    spec: Spec = field(
+        metadata=schema(description="The Spec containing the source midpoints")
+    )
+    region: Region = field(
+        metadata=schema(description="The Region that midpoints will be inside")
+    )
+    check_path_changes: bool = field(
+        default=True,
+        metadata=schema(
+            description="If True path through scan will not be modified by squash"
+        ),
     )
 
-    def keys(self) -> List:
-        return self.spec.keys()
+    def axes(self) -> List:
+        return self.spec.axes()
 
     def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
         dims = self.spec.create_dimensions(bounds, nested)
-        for key_set in self.region.key_sets():
-            # Squash the dimensions together containing these keys
-            matches = [i for i, d in enumerate(dims) if set(d.keys()) & key_set]
-            assert matches, f"No Specs match keys {list(key_set)}"
+        for axis_set in self.region.axis_sets():
+            # Find the start and end index of any dimensions containing these axes
+            matches = [i for i, d in enumerate(dims) if set(d.axes()) & axis_set]
+            assert matches, f"No Specs match axes {list(axis_set)}"
             si, ei = matches[0], matches[-1]
             if si != ei:
-                # Span Specs, squash them
+                # The axis_set spans multiple Dimensions, squash them together
                 # If the spec to be squashed is nested (inside the Mask or outside)
                 # then check the path changes if requested
                 check_path_changes = (nested or si) and self.check_path_changes
                 squashed = squash_dimensions(dims[si : ei + 1], check_path_changes)
                 dims = dims[:si] + [squashed] + dims[ei + 1 :]
-        # Generate masks from the positions showing what's inside
-        masked_dims = [dim.mask(get_mask(self.region, dim.positions)) for dim in dims]
+        # Generate masks from the midpoints showing what's inside
+        masked_dims = [dim.mask(get_mask(self.region, dim.midpoints)) for dim in dims]
         return masked_dims
 
     # *+ bind more tightly than &|^ so without these overrides we
@@ -210,6 +253,7 @@ class Mask(Spec):
         return if_instance_do(other, Region, lambda o: Mask(self.spec, self.region - o))
 
 
+@dataclass
 class Snake(Spec):
     """Run the Spec in reverse on every other iteration when nested inside
     another Spec. Typically created with the ``~`` operator.
@@ -221,12 +265,12 @@ class Snake(Spec):
         spec = Line("y", 1, 3, 3) * ~Line("x", 3, 5, 5)
     """
 
-    spec: Spec = Field(
-        ..., description="The Spec to run in reverse every other iteration"
+    spec: Spec = field(
+        metadata=schema(description="The Spec to run in reverse every other iteration")
     )
 
-    def keys(self) -> List:
-        return self.spec.keys()
+    def axes(self) -> List:
+        return self.spec.axes()
 
     def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
         dims = self.spec.create_dimensions(bounds, nested)
@@ -235,9 +279,10 @@ class Snake(Spec):
         return dims
 
 
+@dataclass
 class Concat(Spec):
     """Concatenate two Specs together, running one after the other. Each Dimension
-    of left and right must contain the same keys.
+    of left and right must contain the same axes.
 
     .. example_spec::
 
@@ -246,17 +291,21 @@ class Concat(Spec):
         spec = Concat(Line("x", 1, 3, 3), Line("x", 4, 5, 5))
     """
 
-    left: Spec = Field(
-        ..., description="The left-hand Spec to Zip, positions will appear earlier"
+    left: Spec = field(
+        metadata=schema(
+            description="The left-hand Spec to Zip, midpoints will appear earlier"
+        )
     )
-    right: Spec = Field(
-        ..., description="The right-hand Spec to Zip, positions will appear later"
+    right: Spec = field(
+        metadata=schema(
+            description="The right-hand Spec to Zip, midpoints will appear later"
+        )
     )
 
-    def keys(self) -> List:
-        left_keys, right_keys = self.left.keys(), self.right.keys()
-        assert left_keys == right_keys, f"Keys {left_keys} != {right_keys}"
-        return left_keys
+    def axes(self) -> List:
+        left_axes, right_axes = self.left.axes(), self.right.axes()
+        assert left_axes == right_axes, f"axes {left_axes} != {right_axes}"
+        return left_axes
 
     def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
         dims_left = self.left.create_dimensions(bounds, nested)
@@ -268,9 +317,10 @@ class Concat(Spec):
         return dimensions
 
 
+@dataclass
 class Squash(Spec):
-    """Squash the Dimensions together of the scan (but not positions) into one
-    long line.
+    """Squash the Dimensions together of the scan (but not the midpoints) into one
+    linear stack.
 
     See Also:
         `why-squash-can-change-path`
@@ -282,13 +332,18 @@ class Squash(Spec):
         spec = Squash(Line("y", 1, 2, 3) * Line("x", 0, 1, 4))
     """
 
-    spec: Spec = Field(..., description="The Spec to squash the dimensions of")
-    check_path_changes: bool = Field(
-        True, description="If True path through scan will not be modified by squash"
+    spec: Spec = field(
+        metadata=schema(description="The Spec to squash the dimensions of")
+    )
+    check_path_changes: bool = field(
+        default=True,
+        metadata=schema(
+            description="If True path through scan will not be modified by squash"
+        ),
     )
 
-    def keys(self) -> List:
-        return self.spec.keys()
+    def axes(self) -> List:
+        return self.spec.axes()
 
     def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
         # TODO: if we squash we explode the size, can we avoid this?
@@ -298,25 +353,28 @@ class Squash(Spec):
 
 
 def _dimensions_from_indexes(
-    func: Callable[[np.ndarray], Dict[Any, np.ndarray]],
-    keys: List,
+    func: Callable[[np.ndarray], Dict[str, np.ndarray]],
+    axes: List,
     num: int,
     bounds: bool,
 ) -> List[Dimension]:
-    positions_calc = func(np.linspace(0.5, num - 0.5, num))
-    positions = {k: positions_calc[k] for k in keys}
+    # Calc num midpoints (fences) from 0.5 .. num - 0.5
+    midpoints_calc = func(np.linspace(0.5, num - 0.5, num))
+    midpoints = {a: midpoints_calc[a] for a in axes}
     if bounds:
+        # Calc num + 1 bounds (posts) from 0 .. num
         bounds_calc = func(np.linspace(0, num, num + 1))
-        lower = {k: bounds_calc[k][:-1] for k in keys}
-        upper = {k: bounds_calc[k][1:] for k in keys}
-        dimension = Dimension(positions, lower, upper)
+        lower = {a: bounds_calc[a][:-1] for a in axes}
+        upper = {a: bounds_calc[a][1:] for a in axes}
+        dimension = Dimension(midpoints, lower, upper)
     else:
-        dimension = Dimension(positions)
+        dimension = Dimension(midpoints)
     return [dimension]
 
 
+@dataclass
 class Line(Spec):
-    """Linearly spaced points in the given key, with first and last points
+    """Linearly spaced points in the given axis, with first and last points
     centred on start and stop.
 
     .. example_spec::
@@ -326,42 +384,46 @@ class Line(Spec):
         spec = Line("x", 1, 2, 5)
     """
 
-    # TODO: are start and stop positions, bounds, or different for fly/step
-    key: Any = Field(..., description="An identifier for what to move")
-    start: float = Field(..., description="Centre point of the first point of the line")
-    stop: float = Field(..., description="Centre point of the last point of the line")
-    num: int = Field(..., ge=1, description="Number of points to produce")
+    axis: str = field(metadata=schema(description="An identifier for what to move"))
+    start: float = field(
+        metadata=schema(description="Centre point of the first point of the line")
+    )
+    stop: float = field(
+        metadata=schema(description="Centre point of the last point of the line")
+    )
+    num: int = field(metadata=schema(min=1, description="Number of points to produce"))
 
-    def keys(self) -> List:
-        return [self.key]
+    def axes(self) -> List:
+        return [self.axis]
 
-    def _line_from_indexes(self, indexes: np.ndarray) -> Dict[Any, np.ndarray]:
+    def _line_from_indexes(self, indexes: np.ndarray) -> Dict[str, np.ndarray]:
         if self.num == 1:
             # Only one point, stop-start gives length of one point
             step = self.stop - self.start
         else:
             # Multiple points, stop-start gives length of num-1 points
             step = (self.stop - self.start) / (self.num - 1)
-        return {self.key: (indexes - 0.5) * step + self.start}
+        # self.start is the first centre point, but we need the lower bound
+        # of the first point as this is where the index array starts
+        first = self.start - step / 2
+        return {self.axis: indexes * step + first}
 
     def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
         return _dimensions_from_indexes(
-            self._line_from_indexes, self.keys(), self.num, bounds
+            self._line_from_indexes, self.axes(), self.num, bounds
         )
 
-    @classmethod
-    @validate_arguments
+    @alternative_constructor
     def bounded(
-        cls,
-        key=key,
-        lower: float = Field(
-            ..., description="Lower bound of the first point of the line"
-        ),
-        upper: float = Field(
-            ..., description="Upper bound of the last point of the line"
-        ),
-        num: int = num,
-    ):
+        axis: Annotated[str, schema(description="An identifier for what to move")],
+        lower: Annotated[
+            float, schema(description="Lower bound of the first point of the line")
+        ],
+        upper: Annotated[
+            float, schema(description="Upper bound of the last point of the line")
+        ],
+        num: Annotated[int, schema(min=1, description="Number of points to produce")],
+    ) -> "Line":
         """Specify a Line by extreme bounds instead of centre points.
 
         .. example_spec::
@@ -378,12 +440,13 @@ class Line(Spec):
         else:
             # Many points, stop will be produced
             stop = upper - half_step
-        return cls(key, start, stop, num)
+        return Line(axis, start, stop, num)
 
 
+@dataclass
 class Static(Spec):
-    """A static point, repeated "num" times, with "key" at "value". Can
-    be used to set key=value at every point in a scan.
+    """A static point, repeated "num" times, with "axis" at "value". Can
+    be used to set axis=value at every point in a scan.
 
     .. example_spec::
 
@@ -392,24 +455,28 @@ class Static(Spec):
         spec = Line("y", 1, 2, 3) + Static("x", 3)
     """
 
-    key: Any = Field(..., description="An identifier for what to move")
-    value: float = Field(..., description="The value at each point")
-    num: int = Field(1, ge=1, description="How many times to repeat this point")
+    axis: str = field(metadata=schema(description="An identifier for what to move"))
+    value: float = field(metadata=schema(description="The value at each point"))
+    num: int = field(
+        default=1,
+        metadata=schema(min=1, description="How many times to repeat this point"),
+    )
 
-    def keys(self) -> List:
-        return [self.key]
+    def axes(self) -> List:
+        return [self.axis]
 
-    def _repeats_from_indexes(self, indexes: np.ndarray) -> Dict[Any, np.ndarray]:
-        return {self.key: np.full(len(indexes), self.value)}
+    def _repeats_from_indexes(self, indexes: np.ndarray) -> Dict[str, np.ndarray]:
+        return {self.axis: np.full(len(indexes), self.value)}
 
     def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
         return _dimensions_from_indexes(
-            self._repeats_from_indexes, self.keys(), self.num, bounds
+            self._repeats_from_indexes, self.axes(), self.num, bounds
         )
 
 
+@dataclass
 class Spiral(Spec):
-    """Archimedean spiral of "x_key" and "y_key", starting at centre point
+    """Archimedean spiral of "x_axis" and "y_axis", starting at centre point
     ("x_start", "y_start") with angle "rotate". Produces "num" points
     in a spiral spanning width of "x_range" and height of "y_range"
 
@@ -420,21 +487,28 @@ class Spiral(Spec):
         spec = Spiral("x", "y", 1, 5, 10, 50, 30)
     """
 
-    x_key: Any = Field(..., description="An identifier for what to move for x")
-    y_key: Any = Field(..., description="An identifier for what to move for y")
+    x_axis: str = field(
+        metadata=schema(description="An identifier for what to move for x")
+    )
+    y_axis: str = field(
+        metadata=schema(description="An identifier for what to move for y")
+    )
     # TODO: do we like these names?
-    x_start: float = Field(..., description="x centre of the spiral")
-    y_start: float = Field(..., description="y centre of the spiral")
-    x_range: float = Field(..., description="x width of the spiral")
-    y_range: float = Field(..., description="y width of the spiral")
-    num: int = Field(..., description="Number of points in the spiral")
-    rotate: float = Field(0.0, description="How much to rotate the angle of the spiral")
+    x_start: float = field(metadata=schema(description="x centre of the spiral"))
+    y_start: float = field(metadata=schema(description="y centre of the spiral"))
+    x_range: float = field(metadata=schema(description="x width of the spiral"))
+    y_range: float = field(metadata=schema(description="y width of the spiral"))
+    num: int = field(metadata=schema(description="Number of points in the spiral"))
+    rotate: float = field(
+        default=0.0,
+        metadata=schema(description="How much to rotate the angle of the spiral"),
+    )
 
-    def keys(self) -> List:
+    def axes(self) -> List:
         # TODO: reversed from __init__ args, a good idea?
-        return [self.y_key, self.x_key]
+        return [self.y_axis, self.x_axis]
 
-    def _spiral_from_indexes(self, indexes: np.ndarray) -> Dict[Any, np.ndarray]:
+    def _spiral_from_indexes(self, indexes: np.ndarray) -> Dict[str, np.ndarray]:
         # simplest spiral equation: r = phi
         # we want point spacing across area to be the same as between rings
         # so: sqrt(area / num) = ring_spacing
@@ -447,28 +521,35 @@ class Spiral(Spec):
         x_scale = self.x_range / diameter
         y_scale = self.y_range / diameter
         return {
-            self.y_key: self.y_start + y_scale * phi * np.cos(phi + self.rotate),
-            self.x_key: self.x_start + x_scale * phi * np.sin(phi + self.rotate),
+            self.y_axis: self.y_start + y_scale * phi * np.cos(phi + self.rotate),
+            self.x_axis: self.x_start + x_scale * phi * np.sin(phi + self.rotate),
         }
 
     def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
         return _dimensions_from_indexes(
-            self._spiral_from_indexes, self.keys(), self.num, bounds
+            self._spiral_from_indexes, self.axes(), self.num, bounds
         )
 
-    @classmethod
-    @validate_arguments
+    @alternative_constructor
     def spaced(
-        cls,
-        x_key: Any = x_key,
-        y_key: Any = y_key,
-        x_start: float = x_start,
-        y_start: float = y_start,
-        radius: float = Field(..., description="radius of the spiral"),
-        dr: float = Field(..., description="difference between each ring"),
-        rotate: float = rotate,
-    ):
-        """Specify a Spiral equally spaced in "x_key" and "y_key" by specifying
+        x_axis: Annotated[
+            str, schema(description="An identifier for what to move for x")
+        ],
+        y_axis: Annotated[
+            str, schema(description="An identifier for what to move for y")
+        ],
+        x_start: Annotated[float, schema(description="x centre of the spiral")],
+        y_start: Annotated[float, schema(description="y centre of the spiral")],
+        radius: Annotated[float, schema(description="radius of the spiral")],
+        dr: Annotated[float, schema(description="difference between each ring")],
+        rotate: Annotated[
+            float,
+            schema(
+                default=0.0, description="How much to rotate the angle of the spiral"
+            ),
+        ] = 0.0,
+    ) -> "Spiral":
+        """Specify a Spiral equally spaced in "x_axis" and "y_axis" by specifying
         the "radius" and difference between each ring of the spiral "dr"
 
         .. example_spec::
@@ -482,16 +563,26 @@ class Spiral(Spec):
         # so: n_rings * 2 * pi = sqrt(4 * pi * num)
         # so: num = n_rings^2 * pi
         n_rings = radius / dr
-        num = n_rings ** 2 * np.pi
-        return cls(x_key, y_key, x_start, y_start, radius * 2, radius * 2, num, rotate)
+        num = int(n_rings ** 2 * np.pi)
+        return Spiral(
+            x_axis, y_axis, x_start, y_start, radius * 2, radius * 2, num, rotate,
+        )
 
 
 #: Can be used as a special key to indicate how long each point should be
 TIME = "TIME"
 
 
-def fly(spec: Spec, duration: float):
-    """Flyscan, zipping TIME=duration for every position
+#: Can be used as a special key to indicate repeats of a whole spec
+REPEAT = "REPEAT"
+
+
+def fly(spec: Spec, duration: float) -> Spec:
+    """Flyscan, zipping TIME=duration for every frame
+
+    Args:
+        spec: The source `Spec` to continuously move
+        duration: How long to spend at each point in the spec
 
     .. example_spec::
 
@@ -502,8 +593,15 @@ def fly(spec: Spec, duration: float):
     return spec + Static(TIME, duration)
 
 
-def step(spec: Spec, duration: float):
-    """Step scan, adding TIME=duration as an inner dimension for every position
+def step(spec: Spec, duration: float, num: int = 1):
+    """Step scan, adding num x TIME=duration as an inner dimension for
+    every midpoint
+
+    Args:
+        spec: The source `Spec` with midpoints to move to and stop
+        duration: The duration of each scan point
+        num: Number of points to produce with given duration at each of point
+            in the spec
 
     .. example_spec::
 
@@ -511,39 +609,20 @@ def step(spec: Spec, duration: float):
 
         spec = step(Line("x", 1, 2, 3), 0.1)
     """
-    return spec * Static(TIME, duration)
+    return spec * Static(TIME, duration, num)
 
 
-class _UnionModifier:
-    # Modifies all Spec subclasses so Spec->Union[all Spec subclasses]
-    def __init__(self):
-        _spec_subclasses = tuple(self._all_subclasses(Spec))
-        _region_subclasses = tuple(self._all_subclasses(Region))
-        self.spec_union = Union[_spec_subclasses]  # type: ignore
-        self.region_union = Union[_region_subclasses]  # type: ignore
+def repeat(spec: Spec, num: int, blend=False):
+    """Repeat spec num times
 
-        for spec in _spec_subclasses + _region_subclasses:
-            for _, field in spec.__fields__.items():
-                if field.type_ is Spec:
-                    # TODO: Is it better to create a new field than modify?
-                    field.type_ = self.spec_union
-                    field.prepare()
-                elif field.type_ is Region:
-                    field.type_ = self.region_union
-                    field.prepare()
-
-    def _all_subclasses(self, cls):
-        return set(cls.__subclasses__()).union(
-            [s for c in cls.__subclasses__() for s in self._all_subclasses(c)]
-        )
-
-
-_modifier = _UnionModifier()
-
-
-def spec_from_dict(d: Dict) -> Spec:
-    return parse_obj_as(_modifier.spec_union, d)  # type: ignore
-
-
-def spec_from_json(text: str) -> Spec:
-    return parse_raw_as(_modifier.spec_union, text)  # type: ignore
+    Args:
+        spec: The source `Spec` that will be iterated
+        num: The number of times to repeat it
+        blend: If True and the slowest dimension of spec is snaked then
+            the end and start of consecutive iterations of Spec will be
+            blended together, leaving no gap
+    """
+    if blend:
+        return Static(REPEAT, num, num) * spec
+    else:
+        return Line(REPEAT, 1, num, num) * spec
